from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def get_email_contacts(key) -> str | None:
    email_contacts = getattr(settings, "EMAIL_CONTACTS", {})
    if not get_email_enabled():
        email_contact = None
        # raise ImproperlyConfigured("Email not enabled. See settings.EMAIL_ENABLED.")
    else:
        if key not in email_contacts:
            raise ImproperlyConfigured(
                f"Key not found. See settings.EMAIL_CONTACTS. Got key=`{key}`."
            )
        email_contact = email_contacts.get(key)
    return email_contact


def get_email_enabled() -> bool:
    return getattr(settings, "EMAIL_ENABLED", False)



from django_audit_fields import ModelAdminAuditFieldsMixin
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin

from edc_model_admin.mixins import (
    ModelAdminFormAutoNumberMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminRedirectOnDeleteMixin,
    ModelAdminReplaceLabelTextMixin,
    TemplatesModelAdminMixin,
)
from edc_notification.modeladmin_mixins import NotificationModelAdminMixin


class PrnModelAdminMixin(
    TemplatesModelAdminMixin,
    ModelAdminNextUrlRedirectMixin,  # add
    NotificationModelAdminMixin,
    ModelAdminFormInstructionsMixin,  # add
    ModelAdminFormAutoNumberMixin,
    ModelAdminRevisionMixin,  # add
    ModelAdminInstitutionMixin,  # add
    ModelAdminRedirectOnDeleteMixin,
    ModelAdminReplaceLabelTextMixin,
    ModelAdminAuditFieldsMixin,
):
    pass



import re

from clinicedc_constants import UUID_PATTERN, YES
from django.db import models

from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils import get_uuid

from ...identifiers import RequisitionIdentifier

human_readable_pattern = r"^[0-9A-Z]{3}\-[0-9A-Z]{4}$"


class RequisitionIdentifierMixin(models.Model):
    requisition_identifier = models.CharField(
        verbose_name="Requisition Id", max_length=50, unique=True
    )

    identifier_prefix = models.CharField(max_length=50, null=True, editable=False, unique=True)

    primary_aliquot_identifier = models.CharField(
        max_length=18, null=True, editable=False, unique=True
    )

    def save(self, *args, **kwargs):
        if not self.requisition_identifier:
            self.requisition_identifier = get_uuid()
        self.protocol_number = self.get_protocol_number()
        self.requisition_identifier = self.get_requisition_identifier()
        super().save(*args, **kwargs)

    @property
    def human_readable_identifier(self):
        """Returns a human readable requisition identifier."""
        x = self.requisition_identifier
        return f"{x[0:3]}-{x[3:7]}"

    def get_protocol_number(self):
        """Returns the protocol number from the field or
        AppConfig.
        """
        protocol_number = self.protocol_number
        if not self.protocol_number:
            protocol_number = ResearchProtocolConfig().protocol_number
        return protocol_number

    def get_requisition_identifier(self):
        """Converts from uuid to a requisition identifier if
        is_drawn == YES and not already a requisition identifier.
        """
        is_uuid = re.match(UUID_PATTERN, self.requisition_identifier)
        if (self.is_drawn == YES or not self.is_drawn) and is_uuid:
            return RequisitionIdentifier(
                protocol_number=self.protocol_number,
                subject_identifier=self.subject_identifier,
                source_model=self._meta.label_lower,
            ).identifier
        return self.requisition_identifier

    class Meta:
        abstract = True



# Generated by Django 6.0 on 2025-05-06 12:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_reportable", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="gradingdata",
            name="age_upper",
            field=models.IntegerField(null=True),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="age_upper_inclusive",
            field=models.BooleanField(null=True),
        ),
        migrations.AlterField(
            model_name="gradingdata",
            name="age_upper_operator",
            field=models.CharField(max_length=15, null=True),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="age_upper",
            field=models.IntegerField(null=True),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="age_upper_inclusive",
            field=models.BooleanField(null=True),
        ),
        migrations.AlterField(
            model_name="normaldata",
            name="age_upper_operator",
            field=models.CharField(max_length=15, null=True),
        ),
    ]



from django.db import models
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords

from ..model_mixins import AddressModelMixin, ContactModelMixin


class Manager(models.Manager):
    use_in_migrations = True


class Supplier(AddressModelMixin, ContactModelMixin, BaseUuidModel):
    supplier_identifier = models.CharField(max_length=36, unique=True, null=True, blank=True)

    name = models.CharField(max_length=255, unique=True)

    contact = models.CharField(max_length=255, default="", blank=True)

    supplier_datetime = models.DateTimeField(default=timezone.now)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.name[:35]

    def save(self, *args, **kwargs):
        if not self.supplier_identifier:
            next_id = get_next_value(self._meta.label_lower)
            self.supplier_identifier = f"{next_id:06d}"
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Supplier"
        verbose_name_plural = "Suppliers"



from edc_auth.constants import AUDITOR_ROLE, CLINICIAN_ROLE, NURSE_ROLE
from edc_auth.site_auths import site_auths

REPORTABLE_SUPER = "REPORTABLE"
REPORTABLE_VIEW = "REPORTABLE_VIEW"


reportable_codenames = [
    "edc_reportable.add_normaldata",
    "edc_reportable.change_normaldata",
    "edc_reportable.view_normaldata",
    "edc_reportable.delete_normaldata",
    "edc_reportable.add_gradingdata",
    "edc_reportable.change_gradingdata",
    "edc_reportable.view_gradingdata",
    "edc_reportable.delete_gradingdata",
]
site_auths.add_group(*reportable_codenames, name=REPORTABLE_SUPER)
site_auths.add_group(
    *[code for code in reportable_codenames if "view" in code], name=REPORTABLE_VIEW
)

site_auths.update_role(REPORTABLE_VIEW, name=CLINICIAN_ROLE)
site_auths.update_role(REPORTABLE_VIEW, name=NURSE_ROLE)
site_auths.update_role(REPORTABLE_VIEW, name=AUDITOR_ROLE)



class DocumentStatusModelAdminMixin:
    def get_readonly_fields(self, request, obj=None) -> tuple[str, ...]:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return tuple(set(readonly_fields + ("document_status",)))



from .crf_with_requisition_model_mixin import (
    CrfWithRequisitionModelMixin,
    requisition_fk_options,
)
from .requisition_identifier_mixin import (
    RequisitionIdentifierMixin,
    human_readable_pattern,
)
from .requisition_model_mixin import RequisitionModelMixin
from .requisition_status_mixin import RequisitionStatusMixin



from typing import Protocol


class VisitScheduleFieldsProtocol(Protocol):
    def visit_code(self) -> str: ...

    def visit_code_sequence(self) -> int: ...

    def visit_schedule_name(self) -> str: ...

    def schedule_name(self) -> str: ...



from datetime import datetime
from typing import List, Protocol


class BaseNotificationStub(Protocol):
    display_name: str | None
    email_body_template: str
    email_footer_template: str
    email_from: list[str]
    # email_message_cls = Type[EmailMessage]
    email_subject_template: str
    email_test_body_line: str
    email_to: list[str] | None
    name: str | None
    # sms_client: Type[Client]
    sms_template: str
    sms_test_line: str

    @property
    def default_email_to(self) -> List[str]: ...  # noqa

    def notify(
        self,
        force_notify: bool = None,
        use_email: bool = None,
        use_sms: bool = None,
        email_body_template: str = None,
        **kwargs,
    ) -> bool: ...


class NotificationStub(BaseNotificationStub, Protocol): ...


class NotificationModelStub(BaseNotificationStub, Protocol):
    emailed: str
    emailed_datetime: datetime
    enabled: bool
    model: str



# Generated by Django 4.1.2 on 2023-01-24 15:22
from django.db import migrations
from tqdm import tqdm


def update_new_appt_type(apps, schema_editor):
    appointment_type_model_cls = apps.get_model("edc_appointment.appointmenttype")
    appointment_model_cls = apps.get_model("edc_appointment.appointment")
    appointment_model_cls.skip_history_when_saving = True
    historical_model_cls = apps.get_model("edc_appointment.historicalappointment")
    total = appointment_model_cls.objects.all().count()
    for appointment in tqdm(appointment_model_cls.objects.all(), total=total):
        appt_type = appointment_type_model_cls.objects.get(name=appointment.appt_type_old)
        appointment.appt_type = appt_type
        appointment.save(update_fields=["appt_type"])

    for appointment in tqdm(historical_model_cls.objects.all(), total=total):
        appt_type = appointment_type_model_cls.objects.get(name=appointment.appt_type_old)
        appointment.appt_type = appt_type
        appointment.save(update_fields=["appt_type"])


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0035_rename_appointment_appt_type"),
    ]

    operations = [migrations.RunPython(update_new_appt_type)]



from uuid import UUID

from django.http import HttpResponse
from reportlab.lib.pagesizes import A4, landscape
from reportlab.lib.units import cm

from ..models import Stock
from ..pdf_reports import NumberedCanvas as BaseNumberedCanvas
from ..pdf_reports import StockReport


class NumberedCanvas(BaseNumberedCanvas):
    footer_row_height = 20
    pagesize = landscape(A4)


def print_stock_view(request, session_uuid: UUID | None):
    response = HttpResponse(content_type="application/pdf")
    response["Content-Disposition"] = 'attachment; filename="stock_report.pdf"'
    page = dict(
        rightMargin=1 * cm,
        leftMargin=1 * cm,
        topMargin=1 * cm,
        bottomMargin=1 * cm,
        pagesize=landscape(A4),
    )
    pks = []
    if session_uuid:
        pks = request.session.get(str(session_uuid))
    queryset = Stock.objects.filter(pk__in=pks)
    report = StockReport(
        queryset=queryset,
        request=request,
        footer_row_height=20,
        page=page,
        numbered_canvas=NumberedCanvas,
    )

    report.build(response)
    return response



import pandas as pd
from django.apps import apps as django_apps
from django_pandas.io import read_frame

from edc_appointment.constants import NEW_APPT

from ..utils import convert_dates_from_model


def get_appointments(
    subject_identifiers: list[str] | None = None,
    normalize: bool | None = None,
    localize: bool | None = None,
):
    normalize = True if normalize is None else normalize
    localize = True if localize is None else localize
    model_cls = django_apps.get_model("edc_appointment.appointment")
    opts = {}
    if subject_identifiers:
        opts = dict(subject_identifier__in=subject_identifiers)
    df = read_frame(model_cls.objects.filter(**opts), verbose=False)
    df = convert_dates_from_model(df, model_cls, normalize=normalize, localize=localize)
    df = df.rename(columns={"id": "appointment_id", "site": "site_id"})
    df["visit_code_str"] = df["visit_code"]
    df = df[
        [
            "appointment_id",
            "subject_identifier",
            "appt_datetime",
            "appt_reason",
            "appt_status",
            "appt_timing",
            "appt_type",
            "appt_type_other",
            "site_id",
            "timepoint",
            "timepoint_datetime",
            "visit_code",
            "visit_code_sequence",
            "visit_code_str",
        ]
    ]
    # convert visit_code to float using visit_code_sequence
    df["visit_code"] = df["visit_code"].astype(float)
    df["visit_code_sequence"] = df["visit_code_sequence"].astype(float)
    df["appt_datetime"] = df["appt_datetime"].apply(pd.to_datetime)
    df["visit_code_sequence"] = df["visit_code_sequence"].apply(
        lambda x: x / 10.0 if x > 0.0 else 0.0
    )
    df["visit_code"] = df["visit_code"] + df["visit_code_sequence"]

    # next visit
    df_next = df[df.appt_status == NEW_APPT].copy()
    df_next = df_next.groupby("subject_identifier").agg(
        {"visit_code": "min", "appt_datetime": "min"}
    )
    df_next = df_next.rename(
        columns={"visit_code": "next_visit_code", "appt_datetime": "next_appt_datetime"}
    )
    if localize:
        df["appt_datetime"] = df["appt_datetime"].dt.tz_localize(None)
    if normalize:
        df["appt_datetime"] = df["appt_datetime"].dt.normalize()
    df_next["next_visit_code_str"] = (
        df_next["next_visit_code"].astype("int64").apply(lambda x: str(x))
    )
    df_next = df_next.reset_index()
    df = df.merge(df_next, on="subject_identifier", how="left")
    return df



from django.db import models
from django.db.models.deletion import PROTECT
from django.utils import timezone

from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_model.models import BaseUuidModel

from .action_model_mixin import ActionModelMixin
from .action_type import ActionType


class ReferenceManager(models.Manager):
    def get_by_natural_key(self, action_identifier):
        return self.get(action_identifier=action_identifier)


class Reference(NonUniqueSubjectIdentifierFieldMixin, ActionModelMixin, BaseUuidModel):
    """Model used as a default reference model for simple actions
    not created by another model.

    Note: In almost all cases an action is created by a model. The
    creating model is then the "reference" model.
    """

    action_identifier = models.CharField(max_length=25, unique=True)

    report_datetime = models.DateTimeField(default=timezone.now)

    action_type = models.ForeignKey(
        ActionType, on_delete=PROTECT, related_name="action", verbose_name="Action"
    )

    objects = ReferenceManager()

    def natural_key(self):
        return (self.action_identifier,)

    class Meta(BaseUuidModel.Meta, NonUniqueSubjectIdentifierFieldMixin.Meta):
        indexes = (
            *BaseUuidModel.Meta.indexes,
            *NonUniqueSubjectIdentifierFieldMixin.Meta.indexes,
        )



from .listboard_filter import ListboardFilter


class MetaListboardViewFilters(type):
    def __new__(mcs, name, bases, attrs):
        parents = [b for b in bases if isinstance(b, MetaListboardViewFilters)]
        if not parents:
            attrs.update({"filters": []})
            attrs.update({"default_include_filter": None})
            attrs.update({"default_exclude_filter": None})
            return super().__new__(mcs, name, bases, attrs)
        filters = []
        default_include_filter = None
        default_exclude_filter = None
        for attrname, obj in attrs.items():
            if not attrname.startswith("_"):
                if isinstance(obj, ListboardFilter):
                    obj.name = attrname
                    if obj.default and not obj.exclude_filter:
                        default_include_filter = obj
                    elif obj.default and obj.exclude_filter:
                        default_exclude_filter = obj
                    filters.append(obj)
        filters.sort(key=lambda x: x.position)
        attrs.update({"filters": filters})
        attrs.update({"default_include_filter": default_include_filter})
        attrs.update({"default_exclude_filter": default_exclude_filter})
        return super().__new__(mcs, name, bases, attrs)


class ListboardViewFilters(metaclass=MetaListboardViewFilters):
    @property
    def include_filters(self):
        return [f for f in self.filters if f.attr == "f"]

    @property
    def exclude_filters(self):
        return [f for f in self.filters if f.attr == "e"]



from datetime import datetime
from zoneinfo import ZoneInfo

from django.conf import settings
from django.core.exceptions import ValidationError

from edc_utils.text import formatted_datetime

from .research_protocol_config import ResearchProtocolConfig


def date_not_before_study_start(value):
    if value:
        protocol_config = ResearchProtocolConfig()
        dte = datetime(*[*value.timetuple()][0:6], tzinfo=ZoneInfo(settings.TIME_ZONE))
        if dte < protocol_config.study_open_datetime:
            opened = formatted_datetime(protocol_config.study_open_datetime)
            raise ValidationError(
                f"Invalid date. Study opened on {opened}. Got {formatted_datetime(dte)}. "
                f"See edc_protocol.AppConfig."
            )


def datetime_not_before_study_start(value_datetime):
    if value_datetime:
        protocol_config = ResearchProtocolConfig()
        dte = value_datetime
        if dte < protocol_config.study_open_datetime:
            opened = formatted_datetime(protocol_config.study_open_datetime)
            raise ValidationError(
                f"Invalid date/time. Study opened on {opened}. Got {formatted_datetime(dte)}."
                f"See edc_protocol.AppConfig."
            )



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q

from edc_protocol.research_protocol_config import ResearchProtocolConfig

from ..identifiers import AliquotIdentifier


class AliquotCreatorError(Exception):
    pass


class AliquotCreator:
    """A class that creates an aliquot.

    Aliquot is either a primary aliquot (is_primary=True) or
    a child aliquot (is_primary=False).
    """

    aliquot_identifier_cls = AliquotIdentifier
    aliquot_model = "edc_lab.aliquot"

    def __init__(
        self,
        parent_identifier=None,
        identifier_prefix=None,
        requisition_identifier=None,
        subject_identifier=None,
        is_primary=None,
    ):
        self.aliquot_model_cls = django_apps.get_model(self.aliquot_model)
        self.requisition_identifier = requisition_identifier
        self.subject_identifier = subject_identifier
        if not parent_identifier and not is_primary:
            raise AliquotCreatorError(
                "Cannot create child aliquot without parent aliquot identifier. "
                f"Got is_primary={is_primary}."
            )
        self.parent_identifier = parent_identifier
        self.identifier_prefix = (
            identifier_prefix
            or f"{ResearchProtocolConfig().protocol_number}{self.requisition_identifier}"
        )
        if is_primary:
            self.parent_segment = None
        else:
            self.parent_segment = self.parent_identifier[-4:]
        self.is_primary = is_primary

    def create(self, count=None, aliquot_type=None):
        """Returns a newly created or existing non-primary
        aliquot model instance.
        """
        try:
            aliquot = self.aliquot_model_cls.objects.get(
                aliquot_type=aliquot_type,
                count=count,
                parent_identifier=self.parent_identifier,
                is_primary=False,
            )
        except ObjectDoesNotExist:
            aliquot = self._create(
                aliquot_type=aliquot_type,
                count=count,
                parent_identifier=self.parent_identifier,
            )
        return aliquot

    def create_primary(self, aliquot_type=None):
        """Returns an existing or newly created primary
        aliquot model instance.
        """
        try:
            aliquot = self.aliquot_model_cls.objects.get(
                Q(identifier_prefix=self.identifier_prefix)
                | Q(requisition_identifier=self.requisition_identifier),
                is_primary=True,
            )
        except ObjectDoesNotExist:
            aliquot = self._create(count=1, aliquot_type=aliquot_type)
        return aliquot

    def _create(self, parent_identifier=None, count=None, aliquot_type=None):
        """Returns a newly created aliquot."""
        aliquot_identifier_obj = self.aliquot_identifier_cls(
            count=count,
            numeric_code=aliquot_type.numeric_code,
            identifier_prefix=self.identifier_prefix,
            parent_segment=self.parent_segment,
        )

        return self.aliquot_model_cls.objects.create(
            aliquot_identifier=aliquot_identifier_obj.identifier,
            aliquot_type=aliquot_type.name,
            alpha_code=aliquot_type.alpha_code,
            count=count,
            numeric_code=aliquot_type.numeric_code,
            parent_identifier=parent_identifier or aliquot_identifier_obj.identifier,
            identifier_prefix=self.identifier_prefix,
            is_primary=bool(self.is_primary),
            requisition_identifier=self.requisition_identifier or "",
            subject_identifier=self.subject_identifier or "",
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone

if TYPE_CHECKING:
    from ..models import Confirmation, Receive, RepackRequest, Stock


def confirm_stock(
    obj: RepackRequest | Receive | None,
    stock_codes: list[str],
    fk_attr: str | None = None,
    confirmed_by: str | None = None,
    user_created: str | None = None,
) -> tuple[list[str], list[str], list[str]]:
    """Confirm stock instances given a list of stock codes
    and a request/receive pk.

    Called from ConfirmStock view.

    See also: confirm_stock_action
    """
    stock_model_cls: type[Stock] = django_apps.get_model("edc_pharmacy.stock")
    confirmation_model_cls: type[Confirmation] = django_apps.get_model(
        "edc_pharmacy.confirmation"
    )
    stock_codes = [s.strip() for s in stock_codes]
    invalid = []
    confirmed = []
    already_confirmed = []
    opts = {}
    if obj:
        opts = {fk_attr: obj.id}
    for stock_code in stock_codes:
        try:
            stock = stock_model_cls.objects.get(code=stock_code, **opts)
        except ObjectDoesNotExist:
            invalid.append(stock_code)
        else:
            try:
                confirmation_model_cls.objects.get(stock=stock)
            except ObjectDoesNotExist:
                confirmation_model_cls.objects.create(
                    stock=stock,
                    confirmed_datetime=timezone.now(),
                    confirmed_by=confirmed_by or user_created,
                )
                confirmed.append(stock.code)
            else:
                already_confirmed.append(stock.code)
    return confirmed, already_confirmed, invalid


__all__ = ["confirm_stock"]



# Generated by Django 5.2.6 on 2025-09-17 16:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_export", "0021_alter_datarequest_revision_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="datarequest",
            name="description",
            field=models.TextField(default=""),
        ),
        migrations.AlterField(
            model_name="datarequest",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="datarequest",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="datarequest",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="datarequest",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="datarequest",
            name="models",
            field=models.TextField(
                default="",
                help_text='List one table per line, no commas. Use "label lower" format.',
            ),
        ),
        migrations.AlterField(
            model_name="datarequesthistory",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="datarequesthistory",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="datarequesthistory",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="datarequesthistory",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="datarequesthistory",
            name="summary",
            field=models.TextField(default=""),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="exportreceipt",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="exportreceipt",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="exportreceipt",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="exportreceipt",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="filehistory",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="filehistory",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="filehistory",
            name="export_uuid_list",
            field=models.TextField(
                default="",
                help_text="list of export_uuid's of model app_label.model_name",
            ),
        ),
        migrations.AlterField(
            model_name="filehistory",
            name="file_contents",
            field=models.TextField(
                default="", help_text="save contents of file as a list of rows"
            ),
        ),
        migrations.AlterField(
            model_name="filehistory",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="filehistory",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="filehistory",
            name="notification_plan_name",
            field=models.CharField(default="", max_length=50),
        ),
        migrations.AlterField(
            model_name="filehistory",
            name="pk_list",
            field=models.TextField(
                default="", help_text="list of pk's of model app_label.model_name"
            ),
        ),
        migrations.AlterField(
            model_name="historicaldatarequest",
            name="description",
            field=models.TextField(default=""),
        ),
        migrations.AlterField(
            model_name="historicaldatarequest",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicaldatarequest",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicaldatarequest",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicaldatarequest",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicaldatarequest",
            name="models",
            field=models.TextField(
                default="",
                help_text='List one table per line, no commas. Use "label lower" format.',
            ),
        ),
        migrations.AlterField(
            model_name="historicalexportreceipt",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalexportreceipt",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalexportreceipt",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalexportreceipt",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalfilehistory",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalfilehistory",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalfilehistory",
            name="export_uuid_list",
            field=models.TextField(
                default="",
                help_text="list of export_uuid's of model app_label.model_name",
            ),
        ),
        migrations.AlterField(
            model_name="historicalfilehistory",
            name="file_contents",
            field=models.TextField(
                default="", help_text="save contents of file as a list of rows"
            ),
        ),
        migrations.AlterField(
            model_name="historicalfilehistory",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalfilehistory",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalfilehistory",
            name="notification_plan_name",
            field=models.CharField(default="", max_length=50),
        ),
        migrations.AlterField(
            model_name="historicalfilehistory",
            name="pk_list",
            field=models.TextField(
                default="", help_text="list of pk's of model app_label.model_name"
            ),
        ),
        migrations.AlterField(
            model_name="historicalobjecthistory",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalobjecthistory",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalobjecthistory",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalobjecthistory",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalplan",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalplan",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalplan",
            name="excluded_field_names",
            field=models.TextField(blank=True, default="", max_length=500),
        ),
        migrations.AlterField(
            model_name="historicalplan",
            name="field_names",
            field=models.TextField(blank=True, default="", max_length=500),
        ),
        migrations.AlterField(
            model_name="historicalplan",
            name="header_labels",
            field=models.TextField(blank=True, default="", max_length=500),
        ),
        migrations.AlterField(
            model_name="historicalplan",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalplan",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalplan",
            name="lookups",
            field=models.TextField(blank=True, default="", max_length=500),
        ),
        migrations.AlterField(
            model_name="objecthistory",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="objecthistory",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="objecthistory",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="objecthistory",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="excluded_field_names",
            field=models.TextField(blank=True, default="", max_length=500),
        ),
        migrations.AlterField(
            model_name="plan",
            name="field_names",
            field=models.TextField(blank=True, default="", max_length=500),
        ),
        migrations.AlterField(
            model_name="plan",
            name="header_labels",
            field=models.TextField(blank=True, default="", max_length=500),
        ),
        migrations.AlterField(
            model_name="plan",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="lookups",
            field=models.TextField(blank=True, default="", max_length=500),
        ),
        migrations.AlterField(
            model_name="uploadexportreceiptfile",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="uploadexportreceiptfile",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="uploadexportreceiptfile",
            name="errors",
            field=models.TextField(default="", editable=False),
        ),
        migrations.AlterField(
            model_name="uploadexportreceiptfile",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="uploadexportreceiptfile",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
    ]



# Generated by Django 5.2.4 on 2025-09-10 03:24

import django_revision.revision_field
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_sites", "0008_delete_edcsite_remove_siteprofile_description"),
    ]

    operations = [
        migrations.AlterField(
            model_name="edcpermissions",
            name="revision",
            field=django_revision.revision_field.RevisionField(
                blank=True,
                editable=False,
                help_text="System field. From git repository (tag:branch:commit), project metadata, project toml, project VERSION, or settings.",
                max_length=75,
                null=True,
                verbose_name="Revision",
            ),
        ),
    ]



from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.managers import CurrentSiteManager
from edc_sites.model_mixins import SiteModelMixin

from ..model_mixins import HealthFacilityModelMixin, Manager


class HealthFacility(SiteModelMixin, HealthFacilityModelMixin, BaseUuidModel):
    objects = Manager()
    on_site = CurrentSiteManager()
    history = HistoricalRecords()

    class Meta(SiteModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Health Facility"
        verbose_name_plural = "Health Facilities"



# Generated by Django 4.2.7 on 2023-12-05 20:02

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0034_issue"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="datadictionary",
            index=models.Index(
                fields=["model", "number", "prompt"],
                name="edc_data_ma_model_92b469_idx",
            ),
        ),
    ]



from typing import Any

from clinicedc_constants import NEVER, OTHER, YES
from django import forms


class MedicationAdherenceFormValidatorMixin:
    def clean(self: Any):
        # super()._clean()
        self.confirm_visual_scores_match()
        self.required_if(
            YES,
            field="pill_count_performed",
            field_required="pill_count",
            field_required_evaluate_as_int=True,
        )
        self.require_m2m_if_missed_any_pills()
        self.missed_pill_reason_other_specify()

    def confirm_visual_scores_match(self: Any) -> None:
        confirmed = self.cleaned_data.get("visual_score_confirmed")
        if confirmed is not None:
            if int(self.cleaned_data.get("visual_score_slider", "0")) != confirmed:
                raise forms.ValidationError(
                    {"visual_score_confirmed": "Does not match visual score above."}
                )

    def require_m2m_if_missed_any_pills(self: Any) -> None:
        if self.cleaned_data.get("last_missed_pill"):
            if self.cleaned_data.get("last_missed_pill") == NEVER:
                self.m2m_not_required("missed_pill_reason")
            else:
                self.m2m_required("missed_pill_reason")

    def missed_pill_reason_other_specify(self: Any) -> None:
        self.m2m_other_specify(
            OTHER,
            m2m_field="missed_pill_reason",
            field_other="other_missed_pill_reason",
        )



from __future__ import annotations

import re
from copy import deepcopy
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from edc_consent.consent_definition import ConsentDefinition
from edc_consent.exceptions import (
    ConsentDefinitionDoesNotExist,
    ConsentDefinitionValidityPeriodError,
)
from edc_registration import get_registered_subject_model_cls
from edc_sites import site_sites
from edc_sites.single_site import SingleSite
from edc_utils import formatted_date

from ..exceptions import NotOnScheduleError, NotOnScheduleForDateError
from ..site_visit_schedules import site_visit_schedules
from ..subject_schedule import SubjectSchedule
from ..visit import Visit
from .visit_collection import VisitCollection
from .window import Window

if TYPE_CHECKING:
    from edc_appointment.models import Appointment
    from edc_model.models import BaseUuidModel
    from edc_sites.model_mixins import SiteModelMixin
    from edc_visit_tracking.model_mixins import VisitModelMixin as Base

    from ..models import OffSchedule, OnSchedule, SubjectScheduleHistory

    class RelatedVisitModel(SiteModelMixin, Base, BaseUuidModel):
        pass


class ScheduleError(Exception):
    pass


class ScheduleNameError(Exception):
    pass


class AlreadyRegisteredVisit(Exception):  # noqa: N818
    pass


class VisitTimepointError(Exception):
    pass


class Schedule:
    """A class that represents a "schedule" of visits.

    Is contained by a "visit schedule".

    Contains an ordered dictionary of visit instances and the onschedule
    and offschedule models used to get on and off the schedule.
    """

    name_regex = r"[a-z0-9\_\-]+$"
    visit_cls = Visit
    visit_collection_cls: type[VisitCollection] = VisitCollection
    window_cls = Window

    def __init__(
        self,
        name: str,
        onschedule_model: str,
        offschedule_model: str,
        consent_definitions: list[ConsentDefinition] | ConsentDefinition = None,
        loss_to_followup_model: str | None = None,
        appointment_model: str | None = None,
        offstudymedication_model: str | None = None,
        history_model: str | None = None,
        verbose_name: str | None = None,
        sequence: str | None = None,
        base_timepoint: float | Decimal | None = None,
    ):
        self._consent_definitions = None
        if not name or not re.match(r"[a-z0-9_\-]+$", name):
            raise ScheduleNameError(
                f"Invalid name. Got '{name}'. May only contains numbers, "
                "lower case letters and '_'."
            )
        self.name = name

        self.consent_definitions = consent_definitions

        if isinstance(base_timepoint, (float,)):
            base_timepoint = Decimal(str(base_timepoint))
        elif isinstance(base_timepoint, (int,)):
            base_timepoint = Decimal(str(base_timepoint) + ".0")
        self._visits = self.visit_collection_cls()
        self.base_timepoint = base_timepoint or Decimal("0.0")
        self.verbose_name = verbose_name or name
        self.sequence = sequence or name
        self.appointment_model: str = appointment_model or "edc_appointment.appointment"
        self.offschedule_model: str = offschedule_model.lower()
        self.onschedule_model: str = onschedule_model.lower()
        self.loss_to_followup_model = (
            None if loss_to_followup_model is None else loss_to_followup_model.lower()
        )
        self.offstudymedication_model = (
            None if offstudymedication_model is None else offstudymedication_model.lower()
        )
        self.history_model = history_model or "edc_visit_schedule.subjectschedulehistory"

    def __repr__(self):
        return f"Schedule({self.name})"

    def __str__(self):
        return self.name

    @property
    def consent_definitions(self) -> list[ConsentDefinition]:
        return self._consent_definitions

    @consent_definitions.setter
    def consent_definitions(
        self, consent_definitions: list[ConsentDefinition] | ConsentDefinition
    ):
        if isinstance(consent_definitions, (ConsentDefinition,)):
            self._consent_definitions: list[ConsentDefinition] = [consent_definitions]
        else:
            self._consent_definitions: list[ConsentDefinition] = [
                x for x in consent_definitions if x
            ]
        if not self._consent_definitions:
            raise ScheduleError(
                f"ConsentDefinition(s) may not be None. See Schedule `{self}`. "
                f"Got `{consent_definitions}`."
            )
        self._consent_definitions = sorted(self._consent_definitions, key=lambda x: x.version)

    @property
    def visits(self) -> VisitCollection:
        """Returns an unfiltered dictionary of visits in this
        schedule.
        """
        return self._visits

    def visits_for_subject(
        self,
        subject_identifier: str,
        report_datetime: datetime,
        site_id: int | None = None,
        consent_definition: ConsentDefinition | None = None,
    ) -> VisitCollection:
        """Returns a deep copy of visits collection filtered for a
        given consented subject.

        If not consented, returns an empty visit collection.

        Check if the consent definition `extended_by` attribute is
        set. If set, visits/timepoints listed with the extended
        consent definition's `timepoints` are EXCLUDED if the subject
        has NOT completed the consent definition extension model.
        """
        visits = self.visit_collection_cls()
        cdef = self.get_consent_definition(
            report_datetime=report_datetime,
            site=site_sites.get(site_id),
            consent_definition=consent_definition,
        )
        if cdef.get_consent_for(subject_identifier=subject_identifier, site_id=site_id):
            visits = deepcopy(self.visits)
            if cdef.extended_by:
                visits = cdef.extended_by.update_visit_collection(
                    visits,
                    subject_identifier,
                    site_id,
                    original_visit_collection=self.visits,
                )
        return visits

    def add_visit(self, visit=None, **kwargs) -> Visit:
        """Adds a unique visit to the schedule.

        Called when first declaring the schedule at bootup.
        """
        visit = visit or self.visit_cls(**kwargs)
        if visit.timepoint < self.base_timepoint:
            raise VisitTimepointError(
                "Visit timepoint cannot be less than this schedule's base_timepoint. "
                f"See {visit}. Got visit.timepoint={visit.timepoint}."
            )

        for attr in ["code", "title", "timepoint", "rbase"]:
            if getattr(visit, attr) in [getattr(v, attr) for v in self.visits.values()]:
                raise AlreadyRegisteredVisit(
                    f"Visit already registered. Got visit={visit} "
                    f"(offending attr='{attr}'). "
                    f"See schedule '{self}'"
                )
        if not self.visits and visit.timepoint != self.base_timepoint:
            raise VisitTimepointError(
                f"First visit timepoint should be {self.base_timepoint}. Set schedule"
                f".base_timepoint if not using default base_timepoint of 0. "
                f"See {visit}. Got visit.timepoint={visit.timepoint}."
            )
        visit.base_timepoint = self.base_timepoint
        self.visits.update(**{visit.code: visit})
        return visit

    @property
    def field_value(self) -> str:
        return self.name

    def crf_required_at(self, label_lower: str) -> list[str]:
        """Returns a list of visit codes where the CRF is required
        by default.
        """
        visit_codes = []
        for visit_code, visit in self.visits.items():
            if label_lower in [form.model for form in visit.crfs if form.required]:
                visit_codes.append(visit_code)
        return visit_codes

    def requisition_required_at(self, requisition_panel) -> list[str]:
        """Returns a list of visit codes where the requisition is
        required by default.

        A requisition is found by its panel.
        """
        visit_codes = []
        for visit_code, visit in self.visits.items():
            if requisition_panel in [
                form.panel for form in visit.requisitions if form.required
            ]:
                visit_codes.append(visit_code)
        return visit_codes

    def subject(self, subject_identifier: str) -> SubjectSchedule:
        """Returns a SubjectSchedule instance for this subject.

        Note: SubjectSchedule puts a subject on/off schedule by
        updating the on/offschedule models
        """
        visit_schedule, schedule = site_visit_schedules.get_by_onschedule_model(
            self.onschedule_model
        )
        if schedule.name != self.name:
            raise ValueError(
                f"Site visit schedules return the wrong schedule object. "
                f"Expected {self!r} for onschedule_model={self.onschedule_model}. "
                f"Got {schedule!r}."
            )
        return SubjectSchedule(
            subject_identifier, visit_schedule=visit_schedule, schedule=self
        )

    def put_on_schedule(
        self,
        subject_identifier: str,
        onschedule_datetime: datetime | None,
        skip_baseline: bool | None = None,
        skip_get_current_site: bool | None = None,
        # consent_definition: ConsentDefinition = None,
    ) -> None:
        """Puts a subject onto this schedule.

        Wrapper of method SubjectSchedule.put_on_schedule.

        Appointment are created through this pathway.
        """
        rs = get_registered_subject_model_cls().objects.get(
            subject_identifier=subject_identifier
        )
        consent_definition = self.get_consent_definition(
            onschedule_datetime, site_sites.get(rs.site.id)
        )
        formatted_cdefs = [
            f"Cdef(proxy_model='{c.proxy_model}', version='{c.version}')"
            for c in self.consent_definitions
        ]
        if not consent_definition:
            raise ScheduleError(
                f"Consent definition may not be None. Expected one of {formatted_cdefs}."
            )

        if consent_definition not in self.consent_definitions:
            raise ScheduleError(
                "Invalid consent definition for schedule. Expected one of "
                f"{formatted_cdefs}. Got Cdef(proxy_model="
                f"'{consent_definition.proxy_model}', "
                f"version='{consent_definition.version}'). "
            )
        self.subject(subject_identifier).put_on_schedule(
            onschedule_datetime,
            skip_baseline=skip_baseline,
            skip_get_current_site=skip_get_current_site,
            consent_definition=consent_definition,
        )

    def refresh_schedule(
        self,
        subject_identifier: str,
        skip_get_current_site: bool | None = None,
    ) -> None:
        self.subject(subject_identifier).refresh_appointments(
            skip_get_current_site=skip_get_current_site
        )

    def take_off_schedule(
        self, subject_identifier: str, offschedule_datetime: datetime
    ) -> None:
        """Wrapper of method SubjectSchedule.take_off_schedule."""
        self.subject(subject_identifier).take_off_schedule(offschedule_datetime)

    def is_onschedule(self, subject_identifier: str, report_datetime: datetime) -> bool:
        try:
            self.subject(subject_identifier).onschedule_or_raise(
                report_datetime=report_datetime, compare_as_datetimes=True
            )
        except (NotOnScheduleError, NotOnScheduleForDateError):
            return False
        return True

    def datetime_in_window(self, **kwargs):
        return self.window_cls(name=self.name, visits=self.visits, **kwargs).datetime_in_window

    @property
    def onschedule_model_cls(self) -> type[OnSchedule]:
        return django_apps.get_model(self.onschedule_model)

    @property
    def offschedule_model_cls(self) -> type[OffSchedule]:
        return django_apps.get_model(self.offschedule_model)

    @property
    def loss_to_followup_model_cls(self):
        return django_apps.get_model(self.loss_to_followup_model)

    @property
    def ltfu_model_cls(self):
        return self.loss_to_followup_model_cls

    @property
    def history_model_cls(self) -> type[SubjectScheduleHistory]:
        return django_apps.get_model(self.history_model)

    @property
    def appointment_model_cls(self) -> type[Appointment]:
        return django_apps.get_model(self.appointment_model)

    @property
    def visit_model_cls(self) -> type[RelatedVisitModel]:
        return self.appointment_model_cls.related_visit_model_cls()

    def get_consent_definition(
        self,
        report_datetime: datetime,
        site: SingleSite = None,
        consent_definition: ConsentDefinition = None,
    ) -> ConsentDefinition:
        """Returns the ConsentDefinition from this schedule valid for the
        given report date or raises an exception.
        """
        if consent_definition:
            consent_definitions = [
                c for c in self.consent_definitions if c.version == consent_definition.version
            ]
        else:
            consent_definitions = self.consent_definitions
        cdefs = [cdef for cdef in consent_definitions if site in cdef.sites]
        if not cdefs:
            cdefs_as_string = ", ".join(
                [cdef.display_name for cdef in self.consent_definitions]
            )
            raise ConsentDefinitionDoesNotExist(
                "This site does not match any consent definitions for this schedule. "
                f"Consent definitions are: {cdefs_as_string}. Got {site.name}."
            )

        cdefs = sorted(cdefs, key=lambda x: x.version, reverse=True)
        cdef = None
        for cdef in cdefs:
            try:
                cdef.valid_for_datetime_or_raise(report_datetime)
            except ConsentDefinitionValidityPeriodError:
                pass
            else:
                break
        if not cdef:
            date_string = formatted_date(report_datetime)
            cdefs_as_string = ", ".join(
                [cdef.display_name for cdef in self.consent_definitions]
            )
            raise ConsentDefinitionDoesNotExist(
                "Date does not fall within the validity period "
                "of any consent definition "
                f"for this schedule. Consent definitions are: "
                f"{cdefs_as_string}. Got {date_string}."
            )
        return cdef

    def to_dict(self):
        return {k: v.to_dict() for k, v in self.visits.items()}



# see clinicedc_constants
from django.conf import settings

ADVERSE_EVENT_APP_LABEL = getattr(settings, "ADVERSE_EVENT_APP_LABEL", "edc_adverse_event")
ADVERSE_EVENT_ADMIN_SITE = getattr(
    settings, "ADVERSE_EVENT_ADMIN_SITE", "edc_adverse_event_admin"
)
AESI_ACTION = "submit-aesi-report"
AE_FOLLOWUP_ACTION = "submit-ae-followup-report"
AE_INITIAL_ACTION = "submit-initial-ae-report"
AE_SUSAR_ACTION = "submit-ae-susar-report"
AE_TMG_ACTION = "submit-ae-tmg-report"
DEATH_REPORT_ACTION = "submit-death-report"
DEATH_REPORT_TMG_ACTION = "submit-death-report-tmg"
DEATH_REPORT_TMG_SECOND_ACTION = "submit-death-report-tmg-2nd"
HOSPITALIZATION_ACTION = "submit-hospitalization-report"
STUDY_TERMINATION_CONCLUSION_ACTION = "submit-study-termination-conclusion"



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

from .get_related_or_none import get_related_or_none

if TYPE_CHECKING:
    from ..models import Stock


def is_dispensed(stock: Stock | str) -> bool | None:
    if isinstance(stock, str):
        stock_model_cls: type[Stock] = django_apps.get_model("edc_pharmacy.stock")
        try:
            stock = stock_model_cls.objects.get(code=stock)
        except ObjectDoesNotExist:
            stock = None
    if (
        stock
        and get_related_or_none(stock, "confirmation")
        and get_related_or_none(stock, "from_stock")
    ):
        return bool(get_related_or_none(stock, "dispenseitem"))
    return None



from .ae_action_classification import AeActionClassification
from .ae_classification import AeClassification
from .cause_of_death import CauseOfDeath
from .edc_permissions import EdcPermissions
from .sae_reason import SaeReason
from .signals import (
    post_delete_ae_susar,
    update_ae_initial_for_susar,
    update_ae_initial_susar_reported,
    update_ae_notifications_for_tmg_group,
    update_death_notifications_for_tmg_group,
)



# Generated by Django 4.2.7 on 2023-12-05 19:49

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0032_alter_datadictionary_options_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="datadictionary",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="datadictionary",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="dataquery",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="dataquery",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="edcpermissions",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="edcpermissions",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicaldatadictionary",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicaldatadictionary",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicaldataquery",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicaldataquery",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalqueryrule",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalqueryrule",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="queryrule",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="queryrule",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="datadictionary",
            name="prompt",
            field=models.CharField(max_length=500),
        ),
        migrations.AlterField(
            model_name="historicaldatadictionary",
            name="prompt",
            field=models.CharField(max_length=500),
        ),
    ]



from django.contrib import admin
from edc_model.admin import HistoricalModelAdminMixin
from edc_model_admin.history import SimpleHistoryAdmin

from ....admin_site import edc_pharmacy_history_admin
from ....models import Stock
from ...model_admin_mixin import ModelAdminMixin


@admin.register(Stock.history.model, site=edc_pharmacy_history_admin)
class StockHistoryAdmin(ModelAdminMixin, HistoricalModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Stock (History)"
    change_form_title = "Pharmacy: Stock (History)"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    list_display = (
        "code",
        "history_type",
        "formatted_history_date",
        "qty",
        "status",
        "subject_identifier",
        "formatted_in_transit",
        "formatted_confirmed_at_location",
        "formatted_dispensed",
        "revision",
        "history_id",
    )

    search_fields = ("code",)

    @admin.display(description="T", ordering="in_transit", boolean=True)
    def formatted_in_transit(self, obj):
        return obj.in_transit

    @admin.display(description="CL", ordering="confirmed_at_location", boolean=True)
    def formatted_confirmed_at_location(self, obj):
        return obj.confirmed_at_location

    @admin.display(description="D", ordering="dispensed", boolean=True)
    def formatted_dispensed(self, obj):
        return obj.dispensed



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import NoReverseMatch, reverse
from django.utils.html import format_html
from django_audit_fields import ModelAdminAuditFieldsMixin, audit_fieldset_tuple
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin
from edc_model_admin.dashboard import ModelAdminDashboardMixin
from edc_model_admin.mixins import (
    ModelAdminFormAutoNumberMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    TemplatesModelAdminMixin,
)
from edc_sites.admin import SiteModelAdminMixin

from ..forms import NoteForm


class NoteModelAdminMixin(
    SiteModelAdminMixin,
    ModelAdminDashboardMixin,
    ModelAdminAuditFieldsMixin,
    ModelAdminFormAutoNumberMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminRevisionMixin,  # add
    ModelAdminInstitutionMixin,  # add
    ModelAdminNextUrlRedirectMixin,
    TemplatesModelAdminMixin,
):
    """A modeladmin mixin class for the Note model."""

    form = NoteForm
    ordering = ("site", "subject_identifier")

    note_template_name = "edc_qareports/qa_report_note.html"

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "subject_identifier",
                    "report_datetime",
                    "note",
                    "status",
                    "report_model",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "dashboard",
        "subject_identifier",
        "report",
        "status",
        "report_note",
        "report_datetime",
    )

    radio_fields = {"status": admin.VERTICAL}  # noqa: RUF012

    list_filter = (
        "report_datetime",
        "status",
        "report_model",
        "user_created",
        "user_modified",
    )

    search_fields = ("subject_identifier", "report_model")

    @admin.display(description="Report", ordering="report_model")
    def report(self, obj=None):
        try:
            app_label, model = obj.report_model_cls._meta.label_lower.split(".")
        except (LookupError, ValueError):
            pass
        else:
            changelist_url = "_".join([app_label, model, "changelist"])
            try:
                # assume admin site naming convention
                url = reverse(f"{app_label}_admin:{changelist_url}")
            except NoReverseMatch:
                # TODO: find the admin site where this model is registered
                pass
            else:
                return format_html(
                    '<a data-toggle="tooltip" title="go to report" href='
                    '"{url}?q={subject_identifier}">{report_model_cls}</a>',
                    url=url,
                    subject_identifier=obj.subject_identifier,
                    report_model_cls=obj.report_model_cls._meta.verbose_name,
                )
        return obj.report_model

    @admin.display(description="QA Note", ordering="note")
    def report_note(self, obj=None):
        context = dict(note=obj.note)
        return render_to_string(self.note_template_name, context)

    # def redirect_url(self, request, obj, post_url_continue=None) -> str | None:
    #     redirect_url = super().redirect_url(request, obj, post_url_continue=post_url_continue)
    #     return f"{redirect_url}?q={obj.subject_identifier}"



account_manager = [
    "auth.add_user",
    "auth.change_permission",
    "auth.change_user",
    "auth.delete_permission",
    "auth.delete_user",
    "auth.view_group",
    "auth.view_permission",
    "auth.view_user",
    "edc_auth.add_userprofile",
    "edc_auth.change_userprofile",
    "edc_auth.delete_userprofile",
    "edc_auth.view_userprofile",
    "edc_auth.add_role",
    "edc_auth.change_role",
    "edc_auth.delete_role",
    "edc_auth.view_role",
]



import calendar

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils import timezone

from edc_model.models import BaseUuidModel
from edc_model_fields.fields import OtherCharField


class HealthFacilityCalendarError(Exception):
    pass


class Manager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, name):
        return self.get(name=name)


class HealthFacilityModelMixin(models.Model):
    """
    Note: Not yet integrated with the Facility object defined on
    visit objects and used to create appointments from the
    visit schedule.

    See also edc_appointment.
    """

    report_datetime = models.DateTimeField(default=timezone.now)

    name = models.CharField(max_length=25, unique=True)

    title = models.CharField(max_length=150, default="", blank=True)

    health_facility_type = models.ForeignKey(
        "edc_facility.HealthFacilityTypes",
        verbose_name="Health facility type",
        on_delete=models.PROTECT,
        related_name="+",
    )

    health_facility_type_other = OtherCharField()

    gps = models.CharField(
        max_length=50,
        default="",
        blank=True,
        help_text="copy and paste directly from google maps",
    )

    latitude = models.FloatField(
        validators=[MinValueValidator(-90), MaxValueValidator(90)],
        null=True,
        blank=True,
        help_text="in degrees. copy and paste directly from google maps",
    )

    longitude = models.FloatField(
        validators=[MinValueValidator(-180), MaxValueValidator(180)],
        null=True,
        blank=True,
        help_text="in degrees. copy and paste directly from google maps",
    )

    mon = models.BooleanField()
    tue = models.BooleanField()
    wed = models.BooleanField()
    thu = models.BooleanField()
    fri = models.BooleanField()
    sat = models.BooleanField()
    sun = models.BooleanField()

    notes = models.TextField(default="", blank=True)

    objects = Manager()

    def __str__(self):
        return f"{self.name.upper()} {self.health_facility_type.display_name}"

    def save(self, *args, **kwargs):
        self.name = self.name.upper()
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.name,)

    @property
    def clinic_days(self) -> list[int]:
        """Using non-ISO numbering where Monday=0."""
        if calendar.firstweekday() != 0:
            raise HealthFacilityCalendarError(
                f"Expected first day of week to be 0. Got {calendar.firstweekday()}."
            )
        days = []
        if self.mon:
            days.append(calendar.MONDAY)
        if self.tue:
            days.append(calendar.TUESDAY)
        if self.wed:
            days.append(calendar.WEDNESDAY)
        if self.thu:
            days.append(calendar.THURSDAY)
        if self.fri:
            days.append(calendar.FRIDAY)
        if self.sat:
            days.append(calendar.SATURDAY)
        if self.sun:
            days.append(calendar.SUNDAY)
        return days

    @property
    def clinic_days_str(self) -> str:
        days = []
        mapping = {k: v for k, v in enumerate(calendar.weekheader(3).split(" "))}
        for day_int in self.clinic_days:
            days.append(mapping.get(day_int))  # noqa: PERF401
        return ",".join(days)

    class Meta(BaseUuidModel.Meta):
        abstract = True
        verbose_name = "Health Facility"
        verbose_name_plural = "Health Facilities"



# Generated by Django 4.2.7 on 2023-12-04 22:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_refusal", "0008_alter_historicalsubjectrefusal_device_created_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="refusalreasons",
            options={
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Refusal Reason",
                "verbose_name_plural": "Refusal Reasons",
            },
        ),
        migrations.AlterModelOptions(
            name="subjectrefusal",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Subject Refusal",
                "verbose_name_plural": "Subject Refusals",
            },
        ),
        migrations.RemoveIndex(
            model_name="refusalreasons",
            name="edc_refusal_id_be47bd_idx",
        ),
        migrations.AddField(
            model_name="historicalsubjectrefusal",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalsubjectrefusal",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="subjectrefusal",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="subjectrefusal",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="refusalreasons",
            name="display_index",
            field=models.IntegerField(
                default=0,
                help_text="Index to control display order if not alphabetical, not required",
                verbose_name="display index",
            ),
        ),
        migrations.AlterField(
            model_name="refusalreasons",
            name="display_name",
            field=models.CharField(
                help_text="(suggest 40 characters max.)",
                max_length=250,
                unique=True,
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="refusalreasons",
            name="name",
            field=models.CharField(
                help_text="This is the stored value, required",
                max_length=250,
                unique=True,
                verbose_name="Stored value",
            ),
        ),
        migrations.AddIndex(
            model_name="refusalreasons",
            index=models.Index(
                fields=["display_index", "display_name"],
                name="edc_refusal_display_22b812_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="subjectrefusal",
            index=models.Index(
                fields=["modified", "created"], name="edc_refusal_modifie_49a6c8_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="subjectrefusal",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_refusal_user_mo_a67c8d_idx",
            ),
        ),
    ]



# Generated by Django 4.2.11 on 2024-06-17 22:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("sites", "0002_alter_domain_unique"),
        ("edc_qareports", "0001_initial"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="ReportNote",
            new_name="QaReportNote",
        ),
        migrations.RemoveConstraint(
            model_name="qareportnote",
            name="edc_qareports_reportnote_name_subj_uniq",
        ),
        migrations.RenameIndex(
            model_name="qareportnote",
            new_name="edc_qarepor_modifie_5f102f_idx",
            old_name="edc_qarepor_modifie_fd8974_idx",
        ),
        migrations.RenameIndex(
            model_name="qareportnote",
            new_name="edc_qarepor_user_mo_6166bf_idx",
            old_name="edc_qarepor_user_mo_9875e1_idx",
        ),
        migrations.AddConstraint(
            model_name="qareportnote",
            constraint=models.UniqueConstraint(
                fields=("report_model", "subject_identifier"),
                name="edc_qareports_qareportnote_name_subj_uniq",
            ),
        ),
    ]



from typing import Any

from clinicedc_constants import EQ, FASTING, MILLIGRAMS_PER_DECILITER, MILLIMOLES_PER_LITER
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from edc_constants.choices import FASTING_CHOICES, YES_NO
from edc_glucose.constants import GLUCOSE_HIGH_READING
from edc_lab.choices import RESULT_QUANTIFIER
from edc_reportable.units import MILLIMOLES_PER_LITER_DISPLAY

from ..model_mixin_factories import reportable_result_model_mixin_factory


class GlucoseModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="glucose",
        verbose_name="Blood Glucose",
        units_choices=(
            (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
            (MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER_DISPLAY),
        ),
        decimal_places=2,
        validators=[MinValueValidator(1.00), MaxValueValidator(GLUCOSE_HIGH_READING)],
    ),
    models.Model,
):
    is_poc = models.CharField(
        verbose_name="Was a point-of-care test used?",
        max_length=15,
        choices=YES_NO,
        default="",
    )
    fasting = models.CharField(
        verbose_name="Was this fasting or non-fasting?",
        max_length=25,
        choices=FASTING_CHOICES,
        default="",
        blank=False,
    )

    glucose_quantifier = models.CharField(
        max_length=10,
        choices=RESULT_QUANTIFIER,
        default=EQ,
    )

    def get_summary_options(self: Any) -> dict:
        opts = super().get_summary_options()
        fasting = self.fasting == FASTING
        opts.update(fasting=fasting)
        return opts

    class Meta:
        abstract = True



# Generated by Django 2.1 on 2018-08-09 03:01

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_locator", "0009_auto_20180708_1954")]

    operations = [
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="tracking_identifier",
            field=models.CharField(db_index=True, max_length=30),
        ),
        migrations.AlterField(
            model_name="subjectlocator",
            name="tracking_identifier",
            field=models.CharField(max_length=30, unique=True),
        ),
    ]



from django.db import models

from edc_search.model_mixins import SearchSlugManager


class SubjectRefusalManager(SearchSlugManager, models.Manager):
    def get_by_natural_key(self, screening_identifier):
        return self.get(screening_identifier=screening_identifier)



from .autocomplete_admin import SiteProxyAdmin, VisitScheduleAdmin
from .medication import (
    AssignmentAdmin,
    DosageGuidelineAdmin,
    FormulationAdmin,
    FormulationTypeAdmin,
    FrequencyUnitsAdmin,
    MedicationAdmin,
    RouteAdmin,
    UnitsAdmin,
)
from .prescription import RxAdmin, RxRefillAdmin
from .reports import StockAvailabilityModelAdmin
from .scan_duplicates_admin import ScanDuplicatesAdmin
from .stock import (
    AllocationAdmin,
    AllocationProxyAdmin,
    ConfirmationAdmin,
    ConfirmationAtLocationHistoryAdmin,
    ConfirmationAtLocationItemHistoryAdmin,
    ContainerAdmin,
    ContainerTypeAdmin,
    DispenseAdmin,
    DispenseItemAdmin,
    LocationAdmin,
    LotAdmin,
    OrderAdmin,
    OrderItemAdmin,
    ProductAdmin,
    ReceiveAdmin,
    ReceiveItemAdmin,
    RequestRepackAdmin,
    StockAdmin,
    StockProxyAdmin,
    StockRequestAdmin,
    StockRequestItemAdmin,
    StockTransferAdmin,
    StorageBinAdmin,
    StorageBinItemAdmin,
    SupplierAdmin,
)



from django.contrib import messages
from django.utils import timezone


def verify_consent(request=None, consent_obj=None):
    if consent_obj.is_verified:
        return None
    consent_obj.is_verified = True
    consent_obj.is_verified_datetime = timezone.now()
    consent_obj.verified_by = request.user.username
    consent_obj.save(update_fields=["is_verified", "is_verified_datetime", "verified_by"])
    return consent_obj


def unverify_consent(consent_obj=None):
    consent_obj.is_verified = False
    consent_obj.is_verified_datetime = None
    consent_obj.verified_by = None
    consent_obj.save(update_fields=["is_verified", "is_verified_datetime", "verified_by"])
    return consent_obj


def flag_as_verified_against_paper(modeladmin, request, queryset, **kwargs):
    """Flags instance as verified against the paper document."""
    for consent_obj in queryset:
        if not consent_obj.open_data_queries.count() > 0:
            if verify_consent(request, consent_obj):
                messages.add_message(
                    request,
                    messages.SUCCESS,
                    f"'{consent_obj._meta.verbose_name}' for ' "
                    f"{consent_obj.subject_identifier}' "
                    f"has been verified against the paper document.",
                )
        else:
            messages.add_message(
                request,
                messages.ERROR,
                f"'{consent_obj._meta.verbose_name}' for ' "
                f"{consent_obj.subject_identifier}' "
                f"has open data queries.",
            )


flag_as_verified_against_paper.short_description = "Verify consent against paper document"


def unflag_as_verified_against_paper(modeladmin, request, queryset, **kwargs):
    """Unflags instance as verified."""
    for consent_obj in queryset:
        unverify_consent(consent_obj)


# noinspection PyTypeHints
unflag_as_verified_against_paper.short_description = "Unverify consent"  # type:ignore



import sys

from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "edc_identifier"
    verbose_name = "Edc Identifier"
    identifier_modulus = 7
    messages_written = False
    include_in_administration_section = True

    def ready(self):
        if not self.messages_written:
            sys.stdout.write(f"Loading {self.verbose_name} ...\n")
            sys.stdout.write(f" * check-digit modulus: {self.identifier_modulus}\n")
            sys.stdout.write(f" Done loading {self.verbose_name}\n")
        self.messages_written = True



from django.utils.translation import gettext as _
from reportlab.graphics.barcode import code128
from reportlab.lib import colors
from reportlab.lib.enums import TA_CENTER, TA_LEFT, TA_RIGHT
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import cm, mm
from reportlab.pdfbase.pdfmetrics import stringWidth
from reportlab.platypus import Paragraph, SimpleDocTemplate, Spacer, Table, TableStyle

from edc_pdf_reports import NumberedCanvas as BaseNumberedCanvas
from edc_pdf_reports import Report
from edc_pdf_reports.flowables import CheckboxFlowable, TextFieldFlowable
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils.date import to_local

from ..models import StockTransfer


class NumberedCanvas(BaseNumberedCanvas):
    footer_row_height = 60


class ManifestReport(Report):
    def __init__(self, stock_transfer: StockTransfer = None, **kwargs):
        self.stock_transfer = stock_transfer
        self.protocol_name = ResearchProtocolConfig().protocol_title
        super().__init__(**kwargs)

    def draw_header(self, canvas, doc):  # noqa: ARG002
        width, height = A4
        canvas.setFontSize(6)
        text_width = stringWidth(self.protocol_name, "Helvetica", 6)
        canvas.drawRightString(width - text_width, height - 20, self.protocol_name.upper())
        canvas.drawString(
            40,
            height - 30,
            (
                _("Stock Transfer Manifest: %(transfer_identifier)s")
                % {"transfer_identifier": self.stock_transfer.transfer_identifier}
            ).upper(),
        )

    @property
    def queryset(self):
        return self.stock_transfer.stocktransferitem_set.all().order_by(
            "stock__allocation__registered_subject__subject_identifier"
        )

    def get_report_story(self, document_template: SimpleDocTemplate = None, **kwargs):  # noqa: ARG002
        story = []

        data = [
            [
                Paragraph(
                    _("Stock Transfer Manifest").upper(),
                    ParagraphStyle(
                        "Title",
                        fontSize=10,
                        spaceAfter=0,
                        alignment=TA_LEFT,
                        fontName="Helvetica-Bold",
                    ),
                ),
                Paragraph(
                    self.protocol_name.upper(),
                    ParagraphStyle(
                        "Title",
                        fontSize=10,
                        spaceAfter=0,
                        alignment=TA_RIGHT,
                        fontName="Helvetica-Bold",
                    ),
                ),
            ],
        ]
        table = Table(data)
        story.append(table)
        story.append(Spacer(0.1 * cm, 0.5 * cm))

        bold_left_style = ParagraphStyle(
            name="line_data_medium",
            alignment=TA_LEFT,
            fontSize=8,
            fontName="Helvetica-Bold",
        )
        bold_right_style = ParagraphStyle(
            name="line_data_medium",
            alignment=TA_RIGHT,
            fontSize=8,
            fontName="Helvetica-Bold",
        )
        left_style = ParagraphStyle(name="line_data_medium", alignment=TA_LEFT, fontSize=8)
        right_style = ParagraphStyle(name="line_data_medium", alignment=TA_RIGHT, fontSize=8)
        from_location = self.stock_transfer.from_location.display_name
        contact_name = self.stock_transfer.from_location.contact_name or ""
        tel = self.stock_transfer.from_location.contact_tel or ""
        email = self.stock_transfer.from_location.contact_email or ""
        timestamp = to_local(self.stock_transfer.transfer_datetime).strftime("%Y-%m-%d")
        data = [
            [
                Paragraph(_("Reference:"), bold_left_style),
                Paragraph(self.stock_transfer.transfer_identifier, left_style),
                Paragraph(_("Contact:"), bold_right_style),
            ],
            [
                Paragraph(_("Date:"), bold_left_style),
                Paragraph(timestamp, left_style),
                Paragraph(contact_name, right_style),
            ],
            [
                Paragraph(_("From:"), bold_left_style),
                Paragraph(from_location, left_style),
                Paragraph(email, right_style),
            ],
            [
                Paragraph(_("To:"), bold_left_style),
                Paragraph(self.stock_transfer.to_location.display_name, left_style),
                Paragraph(tel, right_style),
            ],
            [
                Paragraph("Items:", bold_left_style),
                Paragraph(str(self.queryset.count()), left_style),
                Paragraph("", right_style),
            ],
        ]
        text_width1 = stringWidth(_("Reference"), "Helvetica", 10)
        table = Table(
            data,
            colWidths=(text_width1 * 1.5, None, None),
            rowHeights=(10, 10, 10, 10, 10),
        )
        story.append(table)

        story.append(self.stock_transfer_items_as_table)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(self.signature_line_as_table)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(self.comment_box_as_table)

        return story

    @property
    def stock_transfer_items_as_table(self) -> Table:
        style = ParagraphStyle(
            name="line_data_medium",
            alignment=TA_CENTER,
            fontSize=8,
            textColor=colors.black,
            fontName="Helvetica-Bold",
        )
        data = [
            [
                Paragraph("", style),
                Paragraph("#", style),
                Paragraph(_("Subject"), style),
                Paragraph(_("Code"), style),
                Paragraph(_("Barcode"), style),
                Paragraph(_("Formulation"), style),
                Paragraph(_("Pack"), style),
            ]
        ]
        for index, stock_transfer_item in enumerate(self.queryset):
            barcode = code128.Code128(
                stock_transfer_item.stock.code, barHeight=5 * mm, barWidth=0.7, gap=1.7
            )
            subject_identifier = (
                stock_transfer_item.stock.allocation.registered_subject.subject_identifier
            )
            formulation = stock_transfer_item.stock.product.formulation
            description = f"{formulation.imp_description} "
            style = ParagraphStyle(
                name="line_data", alignment=TA_CENTER, fontSize=8, leading=10
            )
            style_xsmall = ParagraphStyle(
                name="line_data", alignment=TA_CENTER, fontSize=6, leading=8
            )
            data.append(
                [
                    CheckboxFlowable(name=f"checkbox_{index}"),
                    Paragraph(str(index + 1), style),
                    Paragraph(subject_identifier, style),
                    Paragraph(
                        (
                            f"{stock_transfer_item.stock.code[:3]}-"
                            f"{stock_transfer_item.stock.code[3:]}"
                        ),
                        style,
                    ),
                    barcode,
                    Paragraph(description, style_xsmall),
                    Paragraph(str(stock_transfer_item.stock.container), style),
                ]
            )

        table = Table(
            data,
            colWidths=(0.5 * cm, 1 * cm, None, None, None, None, None),
        )
        table.setStyle(
            TableStyle(
                [
                    ("BACKGROUND", (0, 0), (-1, 0), colors.white),
                    ("TEXTCOLOR", (0, 0), (-1, 0), colors.black),
                    ("ALIGN", (0, 0), (-1, -1), "CENTER"),
                    ("FONTNAME", (0, 0), (-1, 0), "Helvetica-Bold"),
                    ("BOTTOMPADDING", (0, 0), (-1, 0), 0),
                    ("BACKGROUND", (0, 1), (-1, -1), colors.white),
                    ("GRID", (0, 0), (-1, -1), 1, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        return table

    @property
    def signature_line_as_table(self) -> Table:
        style = ParagraphStyle(
            name="line_data_medium", alignment=TA_LEFT, fontSize=8, leading=8
        )
        textfield_width = stringWidth("_________________________", "Helvetica", 8)
        data = [
            [
                TextFieldFlowable(
                    name="issued_by_signature",
                    value="",
                    width=textfield_width,
                    height=10,
                    borderWidth=0.5,
                    fontSize=8,
                ),
                TextFieldFlowable(
                    name="received_by_signature",
                    value="",
                    width=textfield_width,
                    height=10,
                    borderWidth=0.5,
                    fontSize=8,
                ),
                TextFieldFlowable(
                    name="received_by",
                    value="",
                    width=textfield_width,
                    height=10,
                    borderWidth=0.5,
                    fontSize=8,
                ),
                TextFieldFlowable(
                    name="received_count",
                    value="",
                    width=textfield_width / 2,
                    height=10,
                    borderWidth=0.5,
                    fontSize=8,
                ),
            ],
            [
                Paragraph(_("Issued by: signature /date"), style=style),
                Paragraph(_("Received by: signature / date"), style=style),
                Paragraph(_("Received by: printed name"), style=style),
                Paragraph(_("Received count"), style=style),
            ],
        ]
        return Table(data, colWidths=(None, None, None, None), rowHeights=(10, 10))

    @property
    def comment_box_as_table(self) -> Table:
        style = ParagraphStyle(
            name="line_data_medium",
            fontSize=8,
            textColor=colors.black,
            fontName="Helvetica-Bold",
        )
        data = [[Paragraph(_("Comment:"), style)]]
        table = Table(data, rowHeights=(75,))
        table.setStyle(
            TableStyle(
                [
                    ("BACKGROUND", (0, 0), (-1, 0), colors.white),
                    ("TEXTCOLOR", (0, 0), (-1, 0), colors.black),
                    ("ALIGN", (0, 0), (-1, -1), "CENTER"),
                    ("FONTNAME", (0, 0), (-1, 0), "Helvetica-Bold"),
                    ("BOTTOMPADDING", (0, 0), (-1, 0), 0),
                    ("BACKGROUND", (0, 1), (-1, -1), colors.white),
                    ("GRID", (0, 0), (-1, -1), 1, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "TOP"),
                ]
            )
        )
        return table



from edc_auth.site_auths import site_auths

from .auth_objects import LABELING, codenames


def update_site_auths():
    site_auths.add_group(*codenames, name=LABELING, no_delete=False)


update_site_auths()



from __future__ import annotations

from clinicedc_constants import NOT_APPLICABLE, YES
from django.db import models

from edc_constants.choices import YES_NO, YES_NO_NA
from edc_model.models import DurationYMDField
from edc_model.validators import date_not_future

from .calculate_date import update_calculated_date


def rx_initial_review_model_mixin_factory(
    fld_prefix: str | None = None, verbose_name_label: str | None = None
):
    fld_prefix = fld_prefix or "rx_init"

    class AbstractModel(models.Model):
        def save(self, *args, **kwargs):
            update_calculated_date(
                self, fld_prefix=fld_prefix, reference_field="report_datetime"
            )
            super().save(*args, **kwargs)

        class Meta:
            abstract = True

    opts = {
        "fld_prefix": fld_prefix,
        f"{fld_prefix}": models.CharField(
            verbose_name=f"Has the patient started {verbose_name_label}?",
            max_length=15,
            choices=YES_NO_NA,
            default=YES,
        ),
        f"{fld_prefix}_date": models.DateField(
            verbose_name=f"Date started {verbose_name_label}",
            validators=[date_not_future],
            null=True,
            blank=True,
            help_text="If possible, provide the exact date here instead of estimating.",
        ),
        f"{fld_prefix}_ago": DurationYMDField(
            verbose_name=(
                f"If date not known, how long ago did the patient start {verbose_name_label}?"
            ),
            null=True,
            blank=True,
            help_text="If possible, provide the exact date above instead of estimating here.",
        ),
        f"{fld_prefix}_calculated_date": models.DateField(
            verbose_name=f"Estimated date started {verbose_name_label}",
            validators=[date_not_future],
            null=True,
            editable=False,
            help_text="Calculated based on response to `rx_init_ago`",
        ),
        f"{fld_prefix}_date_is_estimated": models.CharField(
            verbose_name=f"Was {verbose_name_label} start date estimated?",
            max_length=15,
            choices=YES_NO,
            default=NOT_APPLICABLE,
            editable=False,
        ),
    }

    for name, fld_cls in opts.items():
        AbstractModel.add_to_class(name, fld_cls)

    return AbstractModel



from __future__ import annotations

import socket
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from django.core.exceptions import ValidationError
from django.core.mail.message import EmailMessage
from django.utils import timezone
from edc_notification.utils import get_email_contacts
from edc_protocol.research_protocol_config import ResearchProtocolConfig

if TYPE_CHECKING:
    from django.contrib.auth.models import User


class FilesEmailerError(ValidationError):
    pass


class FilesEmailer:
    def __init__(
        self,
        path_to_files: Path,
        *,
        file_ext: str,
        user: User,
        export_filenames: list[str],
        verbose: bool | None = None,
    ):
        self.path_to_files = path_to_files
        self.file_ext = file_ext
        self.user = user
        self.export_filenames = export_filenames
        self.verbose = verbose

        self.summary = [str(x) for x in self.export_filenames]
        self.summary.sort()
        self.emailed_to = self.user.email

        self.email_files()

        self.emailed_datetime = timezone.now()

    def email_files(self) -> None:
        email_message = self.get_email_message()
        files = [f for f in self.path_to_files.iterdir() if f.suffix == self.file_ext]
        x = 0
        index = 0
        for index, file in enumerate(files):
            email_message.attach_file(str(file))
            x += 1
            if x >= 10:
                email_message.subject = (
                    f"{email_message.subject} (items "
                    f"{index + 2 - x}-{index + 1} of {len(files)})"
                )
                self.send(email_message)
                email_message = self.get_email_message()
                x = 0
        if x > 0:
            email_message.subject = (
                f"{email_message.subject} (items {index + 2 - x}-{index + 1} of {len(files)})"
            )
            self.send(email_message)
        if self.verbose:
            sys.stdout.write(f"\nEmailed export files to {self.user.email}.\n")

    def get_email_message(self) -> EmailMessage:
        body = [
            f"Hello {self.user.first_name or self.user.username}",
            "The data you requested are attached.",
            (
                "An email can contain no more than 10 attached files. If you selected \n"
                "more than 10 tables for export, you will receive more than one email for \n"
                "this request."
            ),
            (
                "Tables with zero records are not exported so the total number of attached \n"
                "files may be fewer than the number of tables you originally selected."
            ),
            (
                "When importing files into your software note that the data are delimited \n"
                'by a pipe, "|",  instead of a comma. You will need to indicate this when you \n'  # noqa
                "open/import the files into Excel, Numbers or whichever software "
                "you are using."
            ),
            "Your request includes the following data:",
            f"{self.summary}",
            "Thanks",
        ]
        return EmailMessage(
            subject=f"{ResearchProtocolConfig().protocol_name.title()} trial data request",
            body="\n\n".join(body),
            from_email=get_email_contacts("data_request"),
            to=[self.user.email],
        )

    @staticmethod
    def send(email_message) -> None:
        try:
            email_message.send()
        except socket.gaierror as e:
            raise FilesEmailerError(
                "Unable to connect to email server.", code="gaierror"
            ) from e



from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.deletion import ProtectedError

from edc_dashboard.view_mixins import EdcViewMixin
from edc_lab import Manifest as ManifestObject
from edc_lab.models import Box, ManifestItem

from ...view_mixins import ManifestViewMixin
from .action_view import ActionView


class ManageManifestViewError(Exception):
    pass


class ManageManifestView(EdcViewMixin, ManifestViewMixin, ActionView):
    post_action_url = "manage_manifest_listboard_url"
    valid_form_actions = ("add_item", "remove_selected_items")

    @property
    def url_kwargs(self):
        return {
            "action_name": self.kwargs.get("action_name"),
            "manifest_identifier": self.manifest_identifier,
        }

    def process_form_action(self, request=None):
        if self.action == "add_item":
            if self.manifest and self.manifest_item_identifier:
                self.add_item()
        elif self.action == "remove_selected_items":
            self.remove_selected_items()

    def remove_selected_items(self):
        """Deletes the selected items, if allowed."""
        if not self.selected_items:
            message = "Nothing to do. No items have been selected."
            messages.warning(self.request, message)
        elif ManifestItem.objects.filter(
            pk__in=self.selected_items, manifest__shipped=True
        ).exists():
            message = "Unable to remove. Some selected items have already been shipped."
            messages.error(self.request, message)
        else:
            try:
                deleted = ManifestItem.objects.filter(
                    pk__in=self.selected_items, manifest__shipped=False
                ).delete()
                message = f"{deleted[0]} items have been removed."
                messages.success(self.request, message)
            except ProtectedError:
                message = "Unable to remove. Manifest is not empty."
                messages.error(self.request, message)

    def add_item(self):
        """Adds the box to the manifest if validated."""
        manifest_object = ManifestObject(manifest=self.manifest, request=self.request)
        manifest_object.add_box(
            box=self.box, manifest_item_identifier=self.manifest_item_identifier
        )

    @property
    def box(self):
        try:
            return Box.objects.get(box_identifier=self.manifest_item_identifier)
        except ObjectDoesNotExist:
            return None



import re

from django.db.models import CharField
from django.forms import RegexField
from django.utils.translation import gettext as _


class InitialsField(CharField):
    description = _("Custom field for a person's initials")

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("editable", True)
        kwargs.setdefault("verbose_name", _("Initials"))
        kwargs.setdefault("max_length", 3)
        kwargs.setdefault("default", "")
        kwargs.setdefault("help_text", _("Type 2-3 letters, all in uppercase and no spaces"))
        CharField.__init__(self, *args, **kwargs)

    def get_internal_type(self):
        return "CharField"

    def formfield(self, **kwargs):
        defaults = {
            "form_class": RegexField,
            "regex": re.compile("^[A-Z]{2,3}$"),
            "max_length": self.max_length,
            "error_messages": {
                "invalid": _(
                    "Enter valid initials. Must be 2-3 letters, all "
                    "in uppercase and no spaces."
                )
            },
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)



from clinicedc_constants import NEVER

MISSED_PILLS = (
    ("today", "today"),
    ("yesterday", "yesterday"),
    ("earlier_this_week", "earlier this week"),
    ("last_week", "last week"),
    ("lt_month_ago", "less than a month ago"),
    ("gt_month_ago", "more than a month ago"),
    (NEVER, "have never missed taking my study pills"),
)



from typing import Any

from edc_lab.constants import SHIPPED

from ...view_mixins import ManifestViewMixin
from .base_listboard_view import BaseListboardView


class ManageManifestListboardView(ManifestViewMixin, BaseListboardView):
    action_name = "manage"
    navbar_selected_item = "manifest"
    form_action_url = "manage_manifest_item_form_action_url"
    listboard_url = "manage_manifest_listboard_url"
    listboard_template = "manage_manifest_listboard_template"
    listboard_view_permission_codename = "edc_lab_dashboard.view_lab_manifest_listboard"
    search_form_url = "manage_manifest_listboard_url"

    listboard_model = "edc_lab.manifestitem"

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        kwargs.update(SHIPPED=SHIPPED, paginator_url_kwargs=self.url_kwargs)
        return super().get_context_data(**kwargs)

    @property
    def url_kwargs(self):
        return {
            "action_name": self.action_name,
            "manifest_identifier": self.manifest_identifier,
        }



from .aliquot_view import AliquotView
from .manage_box_item_view import ManageBoxItemView
from .manage_manifest_view import ManageManifestView
from .manifest_view import ManifestView
from .pack_view import PackView
from .process_view import ProcessView
from .receive_view import ReceiveView
from .requisition_view import RequisitionView
from .verify_box_item_view import VerifyBoxItemView



from django.contrib import admin
from django.contrib.admin.decorators import register
from django_audit_fields.admin import ModelAdminAuditFieldsMixin, audit_fieldset_tuple
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin
from edc_model_admin.history import SimpleHistoryAdmin
from edc_model_admin.mixins import (
    ModelAdminFormAutoNumberMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminInstitutionMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminRedirectOnDeleteMixin,
    TemplatesModelAdminMixin,
)

from ..admin_site import edc_data_manager_admin
from ..forms import QueryRuleForm
from ..models import QueryRule, get_rule_handler_choices
from .actions import (
    copy_query_rule_action,
    toggle_active_flag,
    update_query_rules_action,
)


class QueryRuleModelAdminMixin:
    readonly_fields = ("reference",)

    def formfield_for_choice_field(self, db_field, request, **kwargs):
        if db_field.name == "rule_handler_name":
            kwargs["choices"] = get_rule_handler_choices()
        return super().formfield_for_choice_field(db_field, request, **kwargs)

    @staticmethod
    def form_name(obj=None):
        data_dictionaries = [dd for dd in obj.data_dictionaries.all().order_by("number")]
        return ", ".join(list(set([dd.model_verbose_name for dd in data_dictionaries])))

    @staticmethod
    def questions(obj=None):
        numbers = [str(dd.number) for dd in obj.data_dictionaries.all().order_by("number")]
        return ", ".join(numbers)

    @staticmethod
    def field_names(obj=None):
        fields = [dd.field_name for dd in obj.data_dictionaries.all().order_by("field_name")]
        return ", ".join(fields)

    @staticmethod
    def timepoints(obj=None):
        fields = [v.visit_code for v in obj.visit_schedule.all().order_by("timepoint")]
        return ", ".join(fields)

    @staticmethod
    def query_timing(obj=None):
        return f"{obj.timing} {obj.get_timing_units_display()} from {obj.reference_date}"

    @staticmethod
    def requisition(obj=None):
        return obj.requisition_panel

    @staticmethod
    def rule(obj=None):
        return obj.rule_handler_name


@register(QueryRule, site=edc_data_manager_admin)
class QueryRuleAdmin(
    QueryRuleModelAdminMixin,
    TemplatesModelAdminMixin,
    ModelAdminNextUrlRedirectMixin,  # add
    ModelAdminFormInstructionsMixin,  # add
    ModelAdminFormAutoNumberMixin,
    ModelAdminRevisionMixin,  # add
    ModelAdminInstitutionMixin,  # add
    ModelAdminRedirectOnDeleteMixin,
    ModelAdminAuditFieldsMixin,
    SimpleHistoryAdmin,
):
    form = QueryRuleForm

    actions = (
        update_query_rules_action,
        toggle_active_flag,
        copy_query_rule_action,
    )

    list_display = (
        "title",
        "active",
        "form_name",
        "questions",
        "field_names",
        "requisition",
        "timepoints",
        "query_timing",
        "query_priority",
        "rule",
    )

    list_filter = ("query_priority", "active", "rule_handler_name")

    radio_fields = {  # noqa: RUF012
        "reference_date": admin.VERTICAL,
        "timing_units": admin.VERTICAL,
        "query_priority": admin.VERTICAL,
    }

    autocomplete_fields = [
        "sender",
        "visit_schedule",
        "visit_schedule_exclude",
        "data_dictionaries",
        "requisition_panel",
        "recipients",
    ]

    fieldsets = (
        [
            None,
            {
                "fields": (
                    "title",
                    "data_dictionaries",
                    "requisition_panel",
                    "visit_schedule",
                    "visit_schedule_exclude",
                    "query_text",
                    "query_priority",
                    "recipients",
                    ("timing", "timing_units"),
                    "rule_handler_name",
                    "sender",
                    "active",
                    "comment",
                )
            },
        ],
        ["Reference", {"classes": ("collapse",), "fields": ("reference",)}],
        audit_fieldset_tuple,
    )



from clinicedc_constants import NO, YES
from django.contrib import admin
from django.contrib.admin import SimpleListFilter
from django.db.models import Count
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.safestring import mark_safe
from django_audit_fields import audit_fieldset_tuple
from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils.date import to_local
from rangefilter.filters import DateRangeFilterBuilder

from ...admin_site import edc_pharmacy_admin
from ...constants import CENTRAL_LOCATION
from ...forms import StockTransferForm
from ...models import ConfirmationAtLocationItem, Location, StockTransfer, StockTransferItem
from ..actions import print_transfer_stock_manifest_action, transfer_stock_action
from ..model_admin_mixin import ModelAdminMixin


class LocationListFilterMixin:
    title = "To location"
    parameter_name = "tolocation"

    def lookups(self, request, model_admin):  # noqa: ARG002
        locations = [
            (obj.get("name"), obj.get("display_name"))
            for obj in Location.objects.values("name", "display_name")
            .filter(name=CENTRAL_LOCATION)
            .distinct()
            .order_by("display_name")
        ]
        locations.extend(
            [
                (obj.get("name"), obj.get("display_name"))
                for obj in Location.objects.values("name", "display_name")
                .exclude(name=CENTRAL_LOCATION)
                .distinct()
                .order_by("display_name")
            ]
        )
        return tuple(locations)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            qs = queryset.filter(**{self.parameter_name: self.value()})
        return qs


class ToLocationListFilter(LocationListFilterMixin, SimpleListFilter):
    title = "To location"
    parameter_name = "to_location__name"


class FromLocationListFilter(LocationListFilterMixin, SimpleListFilter):
    title = "From location"
    parameter_name = "from_location__name"


class ConfirmedAtSiteListFilter(SimpleListFilter):
    title = "Confirmed at location"
    parameter_name = "confirmed_at_location"

    def lookups(self, request, model_admin):  # noqa: ARG002
        return (YES, YES), (NO, NO)

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value():
            if self.value() == YES:
                qs = queryset.filter(
                    stocktransferitem__stock__confirmed_at_location=True,
                ).annotate(Count("transfer_identifier"))
            elif self.value() == NO:
                qs = queryset.filter(
                    stocktransferitem__stock__confirmed_at_location=False,
                ).annotate(Count("transfer_identifier"))

        return qs


@admin.register(StockTransfer, site=edc_pharmacy_admin)
class StockTransferAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Stock Transfer"
    change_form_title = "Pharmacy: Stock Transfers"
    change_list_note = "T=Transferred to location or 'in transit', CL=Confirmed at location"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20
    ordering = ("-transfer_identifier",)

    autocomplete_fields = ("from_location", "to_location")
    actions = (transfer_stock_action, print_transfer_stock_manifest_action)

    form = StockTransferForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "transfer_identifier",
                    "transfer_datetime",
                    "from_location",
                    "to_location",
                    "item_count",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "transfer_date",
        "location",
        "n",
        "stock_transfer_item_changelist",
        "stock_transfer_item_confirmed_changelist",
        "stock_transfer_item_unconfirmed_changelist",
        "stock_changelist",
    )

    list_filter = (
        ("transfer_datetime", DateRangeFilterBuilder()),
        FromLocationListFilter,
        ToLocationListFilter,
        ConfirmedAtSiteListFilter,
    )

    search_fields = (
        "id",
        "transfer_identifier",
        "stocktransferitem__stock__code",
        "stocktransferitem__stock__allocation__registered_subject__subject_identifier",
    )

    @admin.display(description="TRANSFER #", ordering="transfer_identifier")
    def identifier(self, obj):
        return obj.transfer_identifier

    @admin.display(description="Transfer date", ordering="transfer_datetime")
    def transfer_date(self, obj):
        return to_local(obj.transfer_datetime).date()

    @admin.display(description="expected", ordering="item_count")
    def n(self, obj):
        if obj.item_count != StockTransferItem.objects.filter(stock_transfer=obj).count():
            return f"{StockTransferItem.objects.filter(stock_transfer=obj).count()}/{obj.item_count}"
        return obj.item_count

    @admin.display(description="Location", ordering="to_location")
    def location(self, obj):
        return mark_safe(f"{obj.from_location}&nbsp;&gt;&gt;&nbsp;{obj.to_location}")  # noqa: S308

    @admin.display(description="T")
    def stock_transfer_item_changelist(self, obj):
        count = StockTransferItem.objects.filter(stock_transfer=obj).count()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stocktransferitem_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label=count, title="Go to stock transfer items")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="CL")
    def stock_transfer_item_confirmed_changelist(self, obj):
        num_confirmed_at_site = ConfirmationAtLocationItem.objects.filter(
            stock_transfer_item__stock_transfer=obj
        ).count()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stocktransferitem_changelist")
        url = f"{url}?q={obj.id}&confirmed_at_site={YES}"
        context = dict(
            url=url,
            label=num_confirmed_at_site,
            title="Items confirmed at site",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="not CL")
    def stock_transfer_item_unconfirmed_changelist(self, obj):
        num = StockTransferItem.objects.filter(
            stock_transfer=obj, stock__confirmed_at_location=False
        ).count()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stocktransferitem_changelist")
        url = f"{url}?q={obj.id}&confirmed_at_site={NO}"
        context = dict(
            url=url,
            label=num,
            title="Items not confirmed at site",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Stock", ordering="stock__code")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label="Stock", title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)



from edc_auth.constants import PII, PII_VIEW
from edc_auth.site_auths import site_auths


def update_site_auths():
    site_auths.update_group(
        "edc_locator.add_subjectlocator",
        "edc_locator.change_subjectlocator",
        "edc_locator.view_historicalsubjectlocator",
        "edc_locator.view_subjectlocator",
        name=PII,
    )

    site_auths.update_group(
        "edc_locator.view_historicalsubjectlocator",
        "edc_locator.view_subjectlocator",
        name=PII_VIEW,
    )
    site_auths.add_pii_model("edc_locator.subjectlocator")


update_site_auths()



from copy import copy

from django.conf import settings
from django.core.exceptions import ValidationError

from .constants import CENTRAL_SERVER, CLIENT, MIDDLEMAN, NODE_SERVER


class DeviceIdError(ValidationError):
    pass


class DeviceRoleError(ValidationError):
    pass


class Device:
    default_central_server_id = "99"
    default_device_id = "99"

    def __init__(
        self,
        device_id=None,
        device_role=None,
        central_server_id=None,
        nodes=None,
        middlemen=None,
        **kwargs,
    ):
        self.central_server_id = central_server_id or self.default_central_server_id
        self.nodes = nodes or []
        self.middlemen = middlemen or []

        if central_server_id in self.nodes:
            raise DeviceIdError(
                f"Central server ID may not be included with node IDs. "
                f"Got {self.central_server_id}, nodes={self.nodes}."
            )
        if central_server_id in self.middlemen:
            raise DeviceIdError(
                f"Central server ID may not be included with middleman IDs. "
                f"Got {self.central_server_id}, middlemen={self.middlemen}."
            )

        self.servers = copy(nodes or [])
        self.servers.append(self.central_server_id)
        self.servers.extend(self.middlemen)

        self.is_central_server = False
        self.is_node_server = False
        self.is_middleman_server = False
        self.is_client = False

        self.is_server = True

        self.device_id = self.get_device_id(device_id)
        if not self.device_id:
            self.device_id = self.default_device_id
        self.device_role = self.get_device_role(device_role)

    def get_device_role(self, device_role=None):
        """Returns the device role.

        *Keywords:
            device_role: a device role usually from AppConfig.

        All values must match; settings.DEVICE_ROLE, if specified,
        AppConfig.device_role, if specified, and the calculated role.
        """

        try:
            settings_device_role = settings.DEVICE_ROLE
        except AttributeError:
            settings_device_role = None
        calculated_role = self.calculated_device_role

        if not (
            calculated_role
            == (device_role or calculated_role)
            == (settings_device_role or calculated_role)
        ):
            raise DeviceRoleError(
                f"Device role conflict. Calculated role as '{calculated_role}' but "
                f"got settings.DEVICE_ID='{settings_device_role}', "
                f"AppConfig.device_role='{device_role}' and device_id='{self.device_id}'",
                code="device_role_conflict",
            )

        return calculated_role

    @property
    def calculated_device_role(self):
        """Returns the device role based on the device ID only."""
        if self.device_id == self.central_server_id:
            role = CENTRAL_SERVER
            self.is_central_server = True
        elif self.device_id in self.nodes:
            role = NODE_SERVER
            self.is_node_server = True
        elif self.device_id in self.middlemen:
            role = MIDDLEMAN
            self.is_middleman_server = True
        else:
            role = CLIENT
            self.is_client = True
            self.is_server = False
        return role

    @staticmethod
    def get_device_id(device_id=None):
        try:
            value = settings.DEVICE_ID
        except AttributeError:
            value = device_id
        else:
            value = value or device_id
        if value and device_id:
            if value != device_id:
                raise DeviceIdError(
                    f"AppConfig.device_id conflicts with settings.DEVICE_ID "
                    f"Got '{device_id}' != '{settings.DEVICE_ID}'",
                    code="device_id_conflict",
                )
        return value



# Generated by Django 6.0 on 2026-01-19 00:53

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0100_auto_20260118_2012"),
    ]

    operations = [
        migrations.RenameField(
            model_name="historicalstock",
            old_name="stored_at_site",
            new_name="stored_at_location",
        ),
        migrations.RenameField(
            model_name="stock",
            old_name="stored_at_site",
            new_name="stored_at_location",
        ),
    ]



from .choices import REPORT_STATUS
from .constants import DEFAULT_BASE_FIELDS
from .utils import (
    InvalidFieldName,
    InvalidFormat,
    duration_to_date,
    estimated_date_from_ago,
    model_exists_or_raise,
)



# Generated by Django 3.2.13 on 2022-08-25 01:51

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_review_dashboard", "0001_initial"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="ReviewDashboard",
            new_name="EdcPermissions",
        ),
        migrations.AlterModelOptions(
            name="edcpermissions",
            options={
                "verbose_name": "Edc Permissions",
                "verbose_name_plural": "Edc Permissions",
            },
        ),
    ]
